/*
 * Copyright (C) 2006 TopCoder Inc., All Rights Reserved.
 */
package com.topcoder.document.index.persistence.impl.db;

import java.io.IOException;
import java.io.Reader;
import java.io.Writer;
import java.sql.Clob;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.text.CollationKey;
import java.text.Collator;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

import com.topcoder.db.connectionfactory.DBConnectionFactory;
import com.topcoder.document.index.CollectionIndex;
import com.topcoder.document.index.DocumentIndex;
import com.topcoder.document.index.persistence.IndexPersistenceException;
import com.topcoder.document.index.persistence.impl.IndexPersistenceHelper;
import com.topcoder.document.index.persistence.impl.PersistenceConfigurationException;
import com.topcoder.document.index.wordsource.WordSourceId;

/**
 * Implements <code>IndexPersistence</code> interface by extending <code>AbstractDBIndexPersistence</code> class. It
 * provides the functionality to persist document indices and collection indices into databases. The source identity
 * instance of the document index must be serializable. In this implementation, the word and position data is stored in
 * a single CLOB column for fast accessing, though somewhat not flexible. The compacted word and position data is
 * formatted as '{word 1}{d}{position 1}{d}{position 2}....{position n}{d}{d}{word 2}....'. '{word 1}' represents the
 * word's source string. '{d}' represents the first delimiter in the document ID. '{position n}' represents the
 * position of the word in the document.<br>
 * The DDL of the tables used by this class is shown as follows:<br>
 *
 * <pre>
 *  CREATE TABLE DOCUMENT_COLLECTION_DOCUMENT_XREF
 *  (
 *  COLLECTION_ID VARCHAR(40),
 *  DOCUMENT_ID VARCHAR(40),
 *  FOREIGN KEY (COLLECTION_ID) REFERENCES DOCUMENT_COLLECTION,
 *  FOREIGN KEY (DOCUMENT_ID) REFERENCES DOCUMENT
 *  )
 *
 *  CREATE TABLE DOCUMENT_COLLECTION
 *  (
 *  COLLECTION_ID VARCHAR(40) PRIMARY KEY
 *  )
 *
 *  CREATE TABLE DOCUMENT
 *  (
 *  DOCUMENT_ID VARCHAR(40) PRIMARY KEY,
 *  LOCALE VARCHAR(30),
 *  SOURCE_IDENTITY BLOB,
 *  CLASS_NAME VARCHAR(255),
 *  USE_COUNT INTEGER
 *  )
 *
 *  CREATE TABLE DELIMITER
 *  (
 *  DOCUMENT_ID VARCHAR(40),
 *  DELIMITER_ENTRY VARCHAR(10),
 *  DELIMITER_SORT_ORDER INTEGER,
 *  FOREIGN KEY (DOCUMENT_ID) REFERENCES DOCUMENT
 *  )
 *
 *  CREATE TABLE WORD_DATA
 *  (
 *  DOCUMENT_ID VARCHAR(40),
 *  WORD_DATA CLOB,
 *  FOREIGN KEY (DOCUMENT_ID) REFERENCES DOCUMENT
 *  )
 * </pre>
 *
 * This class is configurable. The configuration can be seen in <code>AbstractDBIndexPersistence</code> class.<br>
 * This class is stateless. However, its thread-safety depends on its super-class.
 *
 * @author AleaActaEst, TCSDEVELOPER
 * @version 1.0
 * @see AbstractDBIndexPersistence
 */
public class FastAccessDBIndexPersistence extends AbstractDBIndexPersistence {
    /**
     * Represents the SQL statement used to add the words of a document index into the persistence.
     */
    private static final String ADD_DOCUMENT_INDEX_WORD = "INSERT INTO WORD_DATA (DOCUMENT_ID, WORD_DATA)"
        + " VALUES (?, ?)";

    /**
     * Represents the SQL statement used to select the words of a document index from the persistence.
     */
    private static final String SELECT_DOCUMENT_INDEX_WORD = "SELECT WORD_DATA FROM WORD_DATA WHERE DOCUMENT_ID = ?";

    /**
     * Represents the SQL statement used to update the words of a document index into the persistence.
     */
    private static final String UPDATE_DOCUMENT_INDEX_WORD = "UPDATE WORD_DATA SET WORD_DATA = ?"
        + " WHERE DOCUMENT_ID = ?";

    /**
     * Represents an empty string.
     */
    private static final String EMPTY = "";

    /**
     * Represents the array of table names whose records should be deleted when a document index is removed.
     */
    private static final String[] REMOVE_TABLES = new String[] {"WORD_DATA", "DELIMITER", "DOCUMENT"};

    /**
     * Creates a new instance of <code>FastAccessDBIndexPersistence</code> class. The DB connection factory and the
     * connection name are loaded from the given namespace in the configuration.
     *
     * @param namespace the namespace where the configurations are loaded.
     * @throws IllegalArgumentException if <code>namespace</code> is <code>null</code> or empty string.
     * @throws PersistenceConfigurationException if the configurations are not valid; or accessing the configuration
     *             manager fails.
     */
    public FastAccessDBIndexPersistence(String namespace) throws PersistenceConfigurationException {
        super(namespace);
    }

    /**
     * Creates a new instance of <code>FastAccessDBIndexPersistence</code> class. The DB connection factory and the
     * connection name are given. When the connection name is empty string or <code>null</code>, the default
     * connection name is used.
     *
     * @param connectionFactory the DB connection factory used to create database connections.
     * @param connectionName the connection name used to create database connections.
     * @throws IllegalArgumentException if <code>connectionFactory</code> is <code>null</code>.
     */
    public FastAccessDBIndexPersistence(DBConnectionFactory connectionFactory, String connectionName) {
        super(connectionFactory, connectionName);
    }

    /**
     * Adds the given document index to the persistence. The initial collection reference count should be 0. When the
     * document index exists in the persistence, it fails. When the database operation fails, the modifications are
     * rolled back.
     *
     * @param documentIndex the document index to be persisted.
     * @throws IllegalArgumentException if <code>documentIndex</code> is <code>null</code>.
     * @throws IndexPersistenceException if persisting the document index fails; or the document index exists in the
     *             persistence; or the source identity instance is not serializable.
     */
    public void addDocumentIndex(DocumentIndex documentIndex) throws IndexPersistenceException {
        DBPersistenceHelper.validateNotNull(documentIndex, "documentIndex");

        // Get the ID and the data
        WordSourceId sourceId = documentIndex.getWordSourceId();
        String id = IndexPersistenceHelper.generateId(sourceId);

        // Obtain the database connection
        Connection conn = getConnection();

        // Begin the transaction
        beginTransaction(conn);

        try {
            // Add the document source ID to the table DOCUMENT
            addWordSourceId(id, sourceId, conn);

            // Add the record first, since there is no way to obtain the CLOB instance directly
            DBPersistenceHelper.executeSQL(conn, ADD_DOCUMENT_INDEX_WORD, new String[] {id, EMPTY}).close();

            // Retrieve and update the CLOB column
            PreparedStatement statement = DBPersistenceHelper.executeSQL(conn, SELECT_DOCUMENT_INDEX_WORD,
                new String[] {id});

            try {
                // Get the CLOB column for update
                ResultSet rs = statement.getResultSet();
                Clob data;

                rs.next();
                data = rs.getClob(1);

                // Obtain the writer of the CLOB column
                Writer writer;

                try {
                    // This is a bug of Informix JDBC Driver 3.00 JC3
                    writer = data.setCharacterStream(0);
                } catch (SQLException e) {
                    // This is the correct way of JDBC specification.
                    writer = data.setCharacterStream(1);
                }

                // Create the persisting data
                createWordData(documentIndex, writer);

                // Update the CLOB column
                DBPersistenceHelper.executeSQL(conn, UPDATE_DOCUMENT_INDEX_WORD, new Object[] {data, id}).close();
            } finally {
                statement.close();
            }
        } catch (SQLException e) {
            // When error occurs, roll back the transaction
            rollbackTransaction(conn);

            throw new IndexPersistenceException("Adding the document index fails.", e);
        } catch (IOException e) {
            // When error occurs, roll back the transaction
            rollbackTransaction(conn);

            throw new IndexPersistenceException("Adding the word data of the document index fails.", e);
        }

        // When all passes, commit the transaction
        commitTransaction(conn);
    }

    /**
     * Gets the document index from the persistence. The ID of the document is given. When the document is not persisted
     * in the persistence, <code>null</code> should be returned.
     *
     * @param wordSourceId the ID of the document whose index is retrieved.
     * @return the document index of the document with the ID.
     * @throws IllegalArgumentException if <code>wordSourceId</code> is <code>null</code>.
     * @throws IndexPersistenceException if retrieving the document index fails; or the source identity instance is not
     *             serializable.
     */
    public DocumentIndex getDocumentIndex(WordSourceId wordSourceId) throws IndexPersistenceException {
        String id = IndexPersistenceHelper.generateId(wordSourceId);
        DocumentIndex index = null;

        // Obtain the database connection
        Connection conn = getConnection();

        try {
            try {
                // Retrieve the document index
                PreparedStatement statement = DBPersistenceHelper.executeSQL(conn, SELECT_DOCUMENT_INDEX_WORD,
                    new String[] {id});

                try {
                    ResultSet rs = statement.getResultSet();

                    if (rs.next()) {
                        // If the record exists, parse the word data and build the document index
                        index = parseWordData(wordSourceId, rs.getClob(1).getCharacterStream());
                    }
                } finally {
                    statement.close();
                }
            } finally {
                conn.close();
            }
        } catch (SQLException e) {
            throw new IndexPersistenceException("Retrieving the document index fails.", e);
        } catch (IOException e) {
            throw new IndexPersistenceException("Retrieving the word data of the document index fails.", e);
        }

        return index;
    }

    /**
     * Removes a document index from the persistence. The document index can be deleted only when there is no more
     * collection containing the document index. The ID of the document is given. When the document is not persisted, it
     * fails. When the database operation fails, the modifications are rolled back.
     *
     * @param wordSourceId the ID of the document whose index is deleted.
     * @throws IllegalArgumentException if <code>wordSourceId</code> is <code>null</code>.
     * @throws IndexPersistenceException if deleting the document index fails; or the document index is not persisted;
     *             or there is at least one collection containing the document index; or the source identity instance is
     *             not serializable.
     */
    public void removeDocumentIndex(WordSourceId wordSourceId) throws IndexPersistenceException {
        removeDocumentIndexImpl(wordSourceId, REMOVE_TABLES);
    }

    /**
     * Adds the collection index to the persistence. When the collection index exists in the persistence, it fails. When
     * the database operation fails, the modifications are rolled back.
     *
     * @param collectionIndex the collection index to be persisted.
     * @throws IllegalArgumentException if <code>collectionIndex</code> is <code>null</code>.
     * @throws IndexPersistenceException if persisting the collection index fails; or the collection index exists in the
     *             persistence; or the source identity instance of one document index is not serializable.
     */
    public void addCollectionIndex(CollectionIndex collectionIndex) throws IndexPersistenceException {
        addCollectionIndexImpl(collectionIndex);
    }

    /**
     * Gets the collection index from the persistence. The ID of the collection index is given. When the collection
     * index is not persisted, <code>null</code> should be returned.
     *
     * @param collectionId the ID of the collection index to be retrieved.
     * @return the collection index with the given ID.
     * @throws IllegalArgumentException if <code>collectionId</code> is <code>null</code> or empty string.
     * @throws IndexPersistenceException if retrieving the collection index fails.
     */
    public CollectionIndex getCollectionIndex(String collectionId) throws IndexPersistenceException {
        return getCollectionIndexImpl(collectionId);
    }

    /**
     * Removes the collection index from the persistence. The ID of the collection index is given. When the collection
     * index is not persisted, it fails. When the database operation fails, the modifications are rolled back.
     *
     * @param collectionId the ID of the collection index to be deleted.
     * @throws IllegalArgumentException if <code>collectionId</code> is <code>null</code> or empty string.
     * @throws IndexPersistenceException if deleting the collection index fails; or the collection index ID cannot be
     *             found in the persistence.
     */
    public void removeCollectionIndex(String collectionId) throws IndexPersistenceException {
        removeCollectionIndexImpl(collectionId);
    }

    /**
     * Updates the collection index in the persistence. When the collection index ID cannot be found in the persistence,
     * it fails. When the database operation fails, the modifications are rolled back.
     *
     * @param collectionIndex the collection index to be updated.
     * @throws IllegalArgumentException if <code>collectionIndex</code> is <code>null</code>.
     * @throws IndexPersistenceException if updating the collection index fails; or the collection index ID cannot be
     *             found in the persistence; or the source identity instance of one document index is not serializable.
     */
    public void updateCollectionIndex(CollectionIndex collectionIndex) throws IndexPersistenceException {
        updateCollectionIndexImpl(collectionIndex);
    }

    /**
     * Increases the collection reference count of the document index. The document ID is given. When the document index
     * is not persisted, it fails. When the database operation fails, the modifications are rolled back.
     *
     * @param wordSourceId the ID of the document whose collection reference count is increased by 1.
     * @throws IllegalArgumentException if <code>wordSourceId</code> is <code>null</code>.
     * @throws IndexPersistenceException if increasing the collection reference count fails; or the document index is
     *             not persisted; or the source identity instance is not serializable.
     */
    public void increaseDocumentUseCount(WordSourceId wordSourceId) throws IndexPersistenceException {
        increaseDocumentUseCountImpl(wordSourceId);
    }

    /**
     * Decreases the collection reference count of the document index. The document ID is given. When the document index
     * is not persisted, it fails. When the database operation fails, the modifications are rolled back.
     *
     * @param wordSourceId the ID of the document whose collection reference count is decreased by 1.
     * @throws IllegalArgumentException if <code>wordSourceId</code> is <code>null</code>.
     * @throws IndexPersistenceException if decreasing the collection reference count fails; or the document index is
     *             not persisted; or the collection reference count is already 0; or the source identity instance is not
     *             serializable.
     */
    public void decreaseDocumentUseCount(WordSourceId wordSourceId) throws IndexPersistenceException {
        decreaseDocumentUseCountImpl(wordSourceId);
    }

    /**
     * Gets a set of all IDs of documents whose indices are persisted in the persistence.
     *
     * @return a set of all IDs of persisted document indices.
     * @throws IndexPersistenceException if retrieving the document IDs fails.
     */
    public Set getIndexedDocuments() throws IndexPersistenceException {
        return getIndexedDocumentsImpl();
    }

    /**
     * Gets the collection reference count of the given document index. When the document index is not persisted, it
     * fails.
     *
     * @param wordSourceId the ID of the document index whose collection reference count is retrieved.
     * @return the reference count of the document index.
     * @throws IllegalArgumentException if <code>wordSourceId</code> is <code>null</code>.
     * @throws IndexPersistenceException if retrieving the collection reference count fails; or the document index is
     *             not persisted; or the source identity instance is not serializable.
     */
    public int getDocumentUseCount(WordSourceId wordSourceId) throws IndexPersistenceException {
        return getDocumentUseCountImpl(wordSourceId);
    }

    /**
     * Creates the persisting data from the given document index. The persisting data of the document index is formed as
     * '{word 1}{d}{position 1}{d}{position 2}....{position n}{d}{d}{word 2}....'. '{word 1}' represents the word's
     * source string. '{d}' represents the first delimiter in the document ID. '{position n}' represents the position of
     * the word in the document. The writer is closed, even if any error occurs.
     *
     * @param index the document index whose persisting data is created.
     * @param data the writer where the persisting data is written to.
     * @throws IOException if writing the persisting data fails.
     */
    private void createWordData(DocumentIndex index, Writer data) throws IOException {
        try {
            // Get the delimiter
            String delimiter = index.getWordSourceId().getDelimiters()[0];

            // Enumerate every word in the index
            Iterator wordIter = index.getWords().entrySet().iterator();
            boolean firstWord = true;

            while (wordIter.hasNext()) {
                if (firstWord) {
                    // If the word is the first word, no need to write delimiter
                    firstWord = false;
                } else {
                    // Otherwise, write two delimiters
                    data.write(delimiter);
                    data.write(delimiter);
                }

                Entry entry = (Entry) wordIter.next();

                // Add the word into the data
                data.write(((CollationKey) entry.getKey()).getSourceString());

                // Add all positions
                Iterator posIter = ((List) entry.getValue()).iterator();

                while (posIter.hasNext()) {
                    data.write(delimiter);
                    data.write(posIter.next().toString());
                }
            }
        } finally {
            data.close();
        }
    }

    /**
     * Parses the given persisted index data. The persisted index belongs to the given document ID. The persisted data
     * should be in the form described in <code>createWordData</code>.
     *
     * @param id the document ID whose index is parsed.
     * @param data the persisted index data to be parsed.
     * @return a new instance of document index of the document ID parsed from the persisted data.
     * @throws IndexPersistenceException if <code>data</code> is not a valid persisted index data.
     * @throws IOException if reading the persisted data fails.
     */
    private DocumentIndex parseWordData(WordSourceId id, Reader data) throws IOException, IndexPersistenceException {
        // Get the delimiter and the word comparator
        String delimiter = id.getDelimiters()[0];
        Collator collator = Collator.getInstance(id.getSourceLocale());
        Map wordMap = new HashMap();

        // Set the comparator as primary difference
        collator.setStrength(Collator.PRIMARY);

        // Parse the word data. Initially, the next token is a word
        boolean isWord = true;
        ReaderTokenizer tokenizer = new ReaderTokenizer(data, delimiter);
        List positions = null;

        try {
            while (tokenizer.hasToken()) {
                String token = tokenizer.nextToken();

                if (isWord) {
                    // A word is expected, create the entry in the map
                    positions = new ArrayList();

                    if (wordMap.put(collator.getCollationKey(token), positions) != null) {
                        throw new IndexPersistenceException("A word appears twice in the word data.");
                    }

                    // The next one should be a position
                    isWord = false;
                } else if (token.length() == 0) {
                    // Two consecutive delimiters found, the next token should be a word
                    isWord = true;
                } else {
                    // It is a word position, should be an integer
                    try {
                        positions.add(Integer.valueOf(token));
                    } catch (NumberFormatException e) {
                        throw new IndexPersistenceException("A position must be an integer in the word data.", e);
                    }
                }
            }
        } finally {
            data.close();
        }

        try {
            return new DocumentIndex(id, wordMap);
        } catch (IllegalArgumentException e) {
            // When the position is negative, or the position list is empty.
            throw new IndexPersistenceException("The persisted word data is malformed.", e);
        }
    }

    /**
     * Defines a tokenizer which tokenize the content in a reader with one delimiter.
     *
     * @author TCSDEVELOPER
     * @version 1.0
     */
    private static class ReaderTokenizer {
        /**
         * Represents the buffer size when reading the file.
         */
        private static final int BUFFER_SIZE = 1024;

        /**
         * Represents the reader used to retrieve the content to be tokenized.
         */
        private final Reader reader;

        /**
         * Represents the delimiter used to separate tokens.
         */
        private final String delimiter;

        /**
         * Represents the buffer used to hold part of the content in the reader.
         */
        private final StringBuffer buffer = new StringBuffer();

        /**
         * Represents the current position of the buffer. All characters before this position is already processed.
         */
        private int bufferPosition = 0;

        /**
         * Represents a flag indicating whether the reader reaches the end.
         */
        private boolean isEnd = false;

        /**
         * Creates a new instance of <code>ReaderTokenizer</code> class. The reader containing the content and the
         * delimiter separating the tokens are given.
         *
         * @param reader the reader where the content to be tokenized are read.
         * @param delimiter the delimiter used to separate tokens.
         */
        public ReaderTokenizer(Reader reader, String delimiter) {
            this.reader = reader;
            this.delimiter = delimiter;
        }

        /**
         * Get a flag indicating whether there are more tokens.
         *
         * @return <code>true</code> if there are more tokens; <code>false</code> otherwise.
         */
        public boolean hasToken() {
            return !isEnd || (buffer.length() > bufferPosition);
        }

        /**
         * Ensures the length of the buffer in memory. If the file reaches the end, it returns normally.
         *
         * @param length the length of the buffer to be read from the file.
         * @throws IndexPersistenceException if I/O error occurs when reading the file.
         */
        private void ensureBufferLength(int length) throws IndexPersistenceException {
            while (buffer.length() < length) {
                // The buffered length is not enough
                char[] buf = new char[BUFFER_SIZE];

                try {
                    int size = reader.read(buf);

                    if (size < 0) {
                        // Reach the end of the reader
                        isEnd = true;

                        break;
                    }

                    // Put it into the buffer
                    buffer.append(buf, 0, size);
                } catch (IOException e) {
                    throw new IndexPersistenceException("I/O error occurs when reading the file.", e);
                }
            }
        }

        /**
         * Gets the next token in the content using the delimiter.
         *
         * @return the next token using the delimiter.
         * @throws IndexPersistenceException if retrieving the next token fails.
         */
        public String nextToken() throws IndexPersistenceException {
            // The start position of a delimiter
            int wordPos;
            boolean first = true;

            do {
                if (first) {
                    // The first time, try not read.
                    first = false;
                } else {
                    // Read more from the reader
                    ensureBufferLength(buffer.length() + BUFFER_SIZE);
                }

                // Get the first position of a delimiter.
                wordPos = buffer.indexOf(delimiter, bufferPosition);

                // No delimiter can be found, continue to read more if the reader does not reach the end
            } while (!isEnd && (wordPos < 0));

            if (wordPos == -1) {
                // Until end, there is no delimiter. All characters are part of the word.
                wordPos = buffer.length();
            }

            String token = buffer.substring(bufferPosition, wordPos);

            // Remove the delimiter after, if it does not reach the end
            bufferPosition = (wordPos == buffer.length()) ? wordPos : (wordPos + delimiter.length());

            // Clear the buffer
            if (bufferPosition >= BUFFER_SIZE) {
                // When there are too many used characters in the buffer, delete them
                buffer.delete(0, bufferPosition);
                bufferPosition = 0;
            }

            return token;
        }
    }
}











