/*
 * Copyright (C) 2006 TopCoder Inc., All Rights Reserved.
 */
package com.topcoder.document.index.persistence.impl.db;

import java.lang.reflect.InvocationHandler;
import java.lang.reflect.Method;
import java.lang.reflect.Proxy;
import java.sql.Connection;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import com.topcoder.db.connectionfactory.DBConnectionFactory;
import com.topcoder.db.connectionfactory.DBConnectionFactoryImpl;
import com.topcoder.document.index.persistence.IndexPersistenceException;
import com.topcoder.document.index.persistence.TestHelper;
import com.topcoder.document.index.persistence.impl.PersistenceConfigurationException;
import com.topcoder.util.config.ConfigManager;

import junit.framework.Test;
import junit.framework.TestCase;
import junit.framework.TestSuite;

/**
 * Tests functionality and error cases of <code>AbstractDBIndexPersistence</code> class. Since this is an abstract
 * class, its concrete sub-class <code>FastAccessDBIndexPersistence</code> is used. Some error are difficult to be
 * created by a program, such as closing database connection error.
 *
 * @author TCSDEVELOPER
 * @version 1.0
 */
public class AbstractDBIndexPersistenceTestCase extends TestCase implements InvocationHandler {
    /**
     * Represents the configuration manager used in tests.
     */
    private static final ConfigManager CONFIG = ConfigManager.getInstance();

    /**
     * Represents the namespace used in tests.
     */
    private static final String NAMESPACE = "Test";

    /**
     * Represents the DB connection factory used in tests.
     */
    private DBConnectionFactory factory;

    /**
     * Represents the <code>AbstractDBIndexPersistence</code> instance used in tests.
     */
    private AbstractDBIndexPersistence persistence;

    /**
     * Represents the database connection to be closed.
     */
    private Connection connection;

    /**
     * Represents the list of methods invoked in the <code>Connection</code> interface.
     */
    private final List invoked = new ArrayList();

    /**
     * Aggragates all tests in this class.
     *
     * @return test suite aggragating all tests.
     */
    public static Test suite() {
        return new TestSuite(AbstractDBIndexPersistenceTestCase.class);
    }

    /**
     * Sets up the test environment. The all namespaces are removed from the configuration. The DB connection factory
     * namespace is loaded. The DB connection factory instance is created. The test instance is created. The invoked
     * method list is cleared.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    protected void setUp() throws Exception {
        for (Iterator iter = CONFIG.getAllNamespaces(); iter.hasNext();) {
            CONFIG.removeNamespace((String) iter.next());
        }

        CONFIG.add("DBFactory.xml");
        factory = new DBConnectionFactoryImpl("com.topcoder.db.connectionfactory.DBConnectionFactoryImpl");
        persistence = new FastAccessDBIndexPersistence(factory, "connection");
        connection = null;
        invoked.clear();
    }

    /**
     * Cleans up the test environment. The all namespaces are removed from the configuration. The DB connection factory
     * is disposed. The test instance is disposed. The database connection is closed if any.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    protected void tearDown() throws Exception {
        for (Iterator iter = CONFIG.getAllNamespaces(); iter.hasNext();) {
            CONFIG.removeNamespace((String) iter.next());
        }

        factory = null;
        persistence = null;

        if (connection != null) {
            connection.close();
        }
    }

    /**
     * Tests <code>AbstractDBIndexPersistence(String)</code> when the namespace is <code>null</code>. Expect
     * IllegalArgumentException.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testAbstractDBIndexPersistenceStringNull() throws Exception {
        try {
            new FastAccessDBIndexPersistence(null);
            fail("The namespace is null, should throw IllegalArgumentException.");
        } catch (IllegalArgumentException e) {
            // good
        }
    }

    /**
     * Tests <code>AbstractDBIndexPersistence(String)</code> when the namespace is empty string. Expect
     * IllegalArgumentException.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testAbstractDBIndexPersistenceStringEmpty() throws Exception {
        try {
            new FastAccessDBIndexPersistence("  ");
            fail("The namespace is null, should throw IllegalArgumentException.");
        } catch (IllegalArgumentException e) {
            // good
        }
    }

    /**
     * Tests <code>AbstractDBIndexPersistence(String)</code> when the namespace is missing in configuration. Expect
     * PersistenceConfigurationException.
     */
    public void testAbstractDBIndexPersistenceStringNamespaceMissing() {
        try {
            new FastAccessDBIndexPersistence("Missing");
            fail("The namespace is missing in the configuration, should throw PersistenceConfigurationException.");
        } catch (PersistenceConfigurationException e) {
            // good
        }
    }

    /**
     * Tests <code>ABstractDBIndexPersistence(String)</code> when the DB factory namespace is missing. Expect
     * PersistenceConfigurationException.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testAbstractDBIndexPersistenceStringFactoryNamespaceMissing() throws Exception {
        CONFIG.add("DBConfig/FactoryNamespaceMissing.xml");

        try {
            new FastAccessDBIndexPersistence(NAMESPACE);
            fail("The DB factory namespace is missing, should throw PersistenceConfigurationException.");
        } catch (PersistenceConfigurationException e) {
            // good
        }
    }

    /**
     * Tests <code>ABstractDBIndexPersistence(String)</code> when the DB factory instance cannot be created. Expect
     * PersistenceConfigurationException.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testAbstractDBIndexPersistenceStringFactoryError() throws Exception {
        CONFIG.add("DBConfig/FactoryError.xml");

        try {
            new FastAccessDBIndexPersistence(NAMESPACE);
            fail("The DB factory instance cannot be created, should throw PersistenceConfigurationException.");
        } catch (PersistenceConfigurationException e) {
            // good
        }
    }

    /**
     * Tests <code>ABstractDBIndexPersistence(String)</code> when the class is not a DB factory. Expect
     * PersistenceConfigurationException.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testAbstractDBIndexPersistenceStringNotFactory() throws Exception {
        CONFIG.add("DBConfig/NotFactory.xml");

        try {
            new FastAccessDBIndexPersistence(NAMESPACE);
            fail("The class is not a DB factory, should throw PersistenceConfigurationException.");
        } catch (PersistenceConfigurationException e) {
            // good
        }
    }

    /**
     * Tests <code>ABstractDBIndexPersistence(String)</code> when a proper constructor of DB factory cannot be found.
     * Expect PersistenceConfigurationException.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testAbstractDBIndexPersistenceStringNoConstructor() throws Exception {
        CONFIG.add("DBConfig/NoConstructor.xml");

        try {
            new FastAccessDBIndexPersistence(NAMESPACE);
            fail("A proper constructor of DB factory cannot be found, should throw PersistenceConfigurationException.");
        } catch (PersistenceConfigurationException e) {
            // good
        }
    }

    /**
     * Tests <code>ABstractDBIndexPersistence(String)</code> when the DB factory class cannot be found. Expect
     * PersistenceConfigurationException.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testAbstractDBIndexPersistenceStringFactoryNotFound() throws Exception {
        CONFIG.add("DBConfig/FactoryNotFound.xml");

        try {
            new FastAccessDBIndexPersistence(NAMESPACE);
            fail("The DB factory class cannot be found, should throw PersistenceConfigurationException.");
        } catch (PersistenceConfigurationException e) {
            // good
        }
    }

    /**
     * Tests <code>AbstractDBIndexPersistence(String)</code> when the DB factory is not a concrete class. Expect
     * PersistenceConfigurationException.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testAbstractDBIndexPersistenceStringNotClass() throws Exception {
        CONFIG.add("DBConfig/NotClass.xml");

        try {
            new FastAccessDBIndexPersistence(NAMESPACE);
            fail("The DB factory is not a concrete class, should throw PersistenceConfigurationException.");
        } catch (PersistenceConfigurationException e) {
            // good
        }
    }

    /**
     * Tests accuracy of <code>AbstractDBIndexPersistence(String)</code> when all properties are specified. The
     * connection name should be loaded. A DB connection factory instance should be created. These are verified via
     * reflection.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testAbstractDBIndexPersistenceStringAccuracy() throws Exception {
        CONFIG.add("DBConfig/Valid.xml");

        // Create the instance
        persistence = new FastAccessDBIndexPersistence(NAMESPACE);

        // Verify
        assertEquals("The connection name should be loaded from configuration.", "connection", TestHelper.getField(
            AbstractDBIndexPersistence.class, persistence, "connectionName"));
        assertTrue("The DBConnectionFactoryImpl instance should be created.", TestHelper.getField(
            AbstractDBIndexPersistence.class, persistence, "connectionFactory") instanceof DBConnectionFactoryImpl);
    }

    /**
     * Tests accuracy of <code>AbstractDBIndexPersistence(String)</code> when all optional properties are missing. The
     * connection name should be <code>null</code>. A DB connection factory instance should be created. These are
     * verified via reflection.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testAbstractDBIndexPersistenceStringOptionalAccuracy() throws Exception {
        CONFIG.add("DBConfig/Optional.xml");

        // Create the instance
        persistence = new FastAccessDBIndexPersistence(NAMESPACE);

        // Verify
        assertNull("The connection name should be null.", TestHelper.getField(AbstractDBIndexPersistence.class,
            persistence, "connectionName"));
        assertTrue("The DBConnectionFactoryImpl instance should be created.", TestHelper.getField(
            AbstractDBIndexPersistence.class, persistence, "connectionFactory") instanceof DBConnectionFactoryImpl);
    }

    /**
     * Tests accuracy of <code>AbstractDBIndexPersistence(String)</code> when the connection name is empty string. The
     * internal connection name should be <code>null</code>, since default connection name is used. These are
     * verified via reflection.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testAbstractDBIndexPersistenceStringConnectionEmptyAccuracy() throws Exception {
        CONFIG.add("DBConfig/ConnectionNameEmpty.xml");

        // Create the instance
        persistence = new FastAccessDBIndexPersistence(NAMESPACE);

        // Verify
        assertNull("The connection name should be null.", TestHelper.getField(AbstractDBIndexPersistence.class,
            persistence, "connectionName"));
    }

    /**
     * Tests <code>AbstractDBIndexPersistence(DBConnectionFactory, String)</code> when the DB connection factory is
     * <code>null</code>. Expect IllegalArgumentException.
     */
    public void testAbstractDBIndexPersistenceDBConnectionFactoryStringFactoryNull() {
        try {
            new FastAccessDBIndexPersistence(null, "connection");
            fail("The DB connection factory is null, should throw IllegalArgumentException.");
        } catch (IllegalArgumentException e) {
            // good
        }
    }

    /**
     * Tests accuracy of <code>AbstractDBIndexPersistence(DBConnectionFactory, String)</code> when the connection name
     * is a normal string. The internal DB connection factory and the connection name should be set as given in the
     * argument. These are verified via reflection.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testAbstractDBIndexPersistenceDBConnectionFactoryStringNormalAccuracy() throws Exception {
        // Create the persistence instance
        persistence = new FastAccessDBIndexPersistence(factory, "connection");

        // Verify
        assertEquals("The DB connection factory should be set as given.", factory, TestHelper.getField(
            AbstractDBIndexPersistence.class, persistence, "connectionFactory"));
        assertEquals("The connection name should be set as given.", "connection", TestHelper.getField(
            AbstractDBIndexPersistence.class, persistence, "connectionName"));
    }

    /**
     * Tests accuracy of <code>AbstractDBIndexPersistence(DBConnectionFactory, String)</code> when the connection name
     * is <code>null</code>. The internal DB connection factory and the connection name should be set as given in the
     * argument. These are verified via reflection.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testAbstractDBIndexPersistenceDBConnectionFactoryStringNullAccuracy() throws Exception {
        // Create the persistence instance
        persistence = new FastAccessDBIndexPersistence(factory, null);

        // Verify
        assertEquals("The DB connection factory should be set as given.", factory, TestHelper.getField(
            AbstractDBIndexPersistence.class, persistence, "connectionFactory"));
        assertNull("The connection name should be set as given.", TestHelper.getField(AbstractDBIndexPersistence.class,
            persistence, "connectionName"));
    }

    /**
     * Tests accuracy of <code>AbstractDBIndexPersistence(DBConnectionFactory, String)</code> when the connection name
     * is empty string. The internal DB connection factory should be set as given in the argument. The internal
     * connection name should be set as <code>null</code>, since the default connection name should be used. These
     * are verified via reflection.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testAbstractDBIndexPersistenceDBConnectionFactoryStringEmptyAccuracy() throws Exception {
        // Create the persistence instance
        persistence = new FastAccessDBIndexPersistence(factory, "   ");

        // Verify
        assertEquals("The DB connection factory should be set as given.", factory, TestHelper.getField(
            AbstractDBIndexPersistence.class, persistence, "connectionFactory"));
        assertNull("The connection name should be set as null.", TestHelper.getField(AbstractDBIndexPersistence.class,
            persistence, "connectionName"));
    }

    /**
     * Tests <code>getConnection()</code> when the connection cannot be established. Expect IndexPersistenceException.
     */
    public void testGetConnectionError() {
        // Create the persistence instance
        persistence = new FastAccessDBIndexPersistence(factory, "error");

        try {
            persistence.getConnection();
            fail("The connection cannot be established, should throw IndexPersistenceException.");
        } catch (IndexPersistenceException e) {
            // good
        }
    }

    /**
     * Tests accuracy of <code>getConnection()</code> when the default connection should be used. A database
     * connection should be obtained. It should be set as manually commit.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testGetConnectionDefaultAccuracy() throws Exception {
        // Create the persistence instance
        persistence = new FastAccessDBIndexPersistence(factory, "connection");

        // Obtain a database connection
        connection = persistence.getConnection();

        // Verify
        assertFalse("The connection should be manually committed.", connection.getAutoCommit());
    }

    /**
     * Tests <code>beginTransaction(Connection)</code> when the connection is <code>null</code>. Expect
     * IllegalArgumentException.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testBeginTransactionConnectionNull() throws Exception {
        try {
            persistence.beginTransaction(null);
            fail("The connection is null, should throw IllegalArgumentException.");
        } catch (IllegalArgumentException e) {
            // good
        }
    }

    /**
     * Tests <code>commitTransaction(Connection)</code> when the connection is <code>null</code>. Expect
     * IllegalArgumentException.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testCommitTransactionConnectionNull() throws Exception {
        try {
            persistence.commitTransaction(null);
            fail("The connection is null, should throw IllegalArgumentException.");
        } catch (IllegalArgumentException e) {
            // good
        }
    }

    /**
     * Tests <code>commitTransaction(Connection)</code> when the database error occurs. Expect
     * IndexPersistenceException.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testCommitTransactionConnectionDBError() throws Exception {
        connection = persistence.getConnection();

        // Close the connection
        connection.close();

        try {
            persistence.commitTransaction(connection);
            fail("Database error occurs, should throw IndexPersistenceException.");
        } catch (IndexPersistenceException e) {
            // good
        }
    }

    /**
     * Tests accuracy of <code>commitTransaction(Connection)</code>. The transaction should be committed and the
     * connection should be closed. These are verified by a proxy instance.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testCommitTransactionConnectionAccuracy() throws Exception {
        // Commit the transaction and close the connection
        persistence.commitTransaction((Connection) Proxy.newProxyInstance(getClass().getClassLoader(),
            new Class[] {Connection.class}, this));

        // Verify
        List expected = new ArrayList();

        expected.add(Connection.class.getMethod("commit", new Class[0]));
        expected.add(Connection.class.getMethod("close", new Class[0]));

        assertEquals("The invoked methods should be commit and close.", expected, invoked);
    }

    /**
     * Tests <code>rollbackTransaction(Connection)</code> when the connection is <code>null</code>. Expect
     * IllegalArgumentException.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testRollbackTransactionConnectionNull() throws Exception {
        try {
            persistence.rollbackTransaction(null);
            fail("The connection is null, should throw IllegalArgumentException.");
        } catch (IllegalArgumentException e) {
            // good
        }
    }

    /**
     * Tests <code>rollbackTransaction(Connection)</code> when the database error occurs. Expect
     * IndexPersistenceException.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testRollbackTransactionConnectionDBError() throws Exception {
        connection = persistence.getConnection();

        // Close the connection
        connection.close();

        try {
            persistence.rollbackTransaction(connection);
            fail("Database error occurs, should throw IndexPersistenceException.");
        } catch (IndexPersistenceException e) {
            // good
        }
    }

    /**
     * Tests accuracy of <code>rollbackTransaction(Connection)</code>. The transaction should be rolled back and the
     * connection should be closed. These are verified by a proxy instance.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testRollbackTransactionConnectionAccuracy() throws Exception {
        // Commit the transaction and close the connection
        persistence.rollbackTransaction((Connection) Proxy.newProxyInstance(getClass().getClassLoader(),
            new Class[] {Connection.class}, this));

        // Verify
        List expected = new ArrayList();

        expected.add(Connection.class.getMethod("rollback", new Class[0]));
        expected.add(Connection.class.getMethod("close", new Class[0]));

        assertEquals("The invoked methods should be commit and close.", expected, invoked);
    }

    /**
     * Handles the method invocation of a interface. The method invoked is added to the invoked method list.
     *
     * @param proxy the proxy instance whose method is invoked.
     * @param method the method to be invoked.
     * @param args the array of arguments of the method.
     * @return the return value of the method.
     */
    public Object invoke(Object proxy, Method method, Object[] args) {
        invoked.add(method);

        return null;
    }
}








