/*
 * Copyright (C) 2006 TopCoder Inc., All Rights Reserved.
 */
package com.topcoder.document.index.persistence.impl.db;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.text.Collator;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.Set;

import com.topcoder.document.index.CollectionIndex;
import com.topcoder.document.index.DocumentIndex;
import com.topcoder.document.index.persistence.IndexPersistenceException;
import com.topcoder.document.index.persistence.TestHelper;
import com.topcoder.document.index.persistence.impl.IndexPersistenceHelper;
import com.topcoder.document.index.wordsource.WordSourceId;

import junit.framework.TestCase;

/**
 * Tests common functionality and error cases of both <code>FastAccessDBIndexPersistence</code> and
 * <code>SimpleAccessDBIndexPersistence</code>. This is an abstract class. Generally, it tests the collection methods
 * and document usage methods. There are some error cases which cannot be produced by a program, such as database
 * connection error within the execution of a method. Therefore, these cases are not tested here, unless there is some
 * programmable control of the database engine. The error of obtaining a database connection is already tested in the
 * <code>AbstractDBIndexPersistence</code>, it is not tested again here.
 *
 * @author TCSDEVELOPER
 * @version 1.0
 */
public abstract class CommonDBIndexPersistenceTestCase extends TestCase {
    /**
     * Represents the <code>AbstractDBIndexPersistence</code> instance used in tests.
     */
    protected AbstractDBIndexPersistence persistence;

    /**
     * Represents the database connection used to access the persistence.
     */
    protected Connection connection;

    /**
     * Creates a new instance of the DB index persistence used in tests.
     *
     * @return a new instance of the DB index persistence to be tested.
     * @throws Exception if error occurs when creating the index persistence.
     */
    protected abstract AbstractDBIndexPersistence createPersistence() throws Exception;

    /**
     * Clears the persistence related tables.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    private void clearTables() throws Exception {
        DBPersistenceHelper.executeSQL(connection, "DELETE FROM DELIMITER", new Object[0]).close();
        DBPersistenceHelper.executeSQL(connection, "DELETE FROM WORD_DATA", new Object[0]).close();
        DBPersistenceHelper.executeSQL(connection, "DELETE FROM WORD_POSITION", new Object[0]).close();
        DBPersistenceHelper.executeSQL(connection, "DELETE FROM WORD", new Object[0]).close();
        DBPersistenceHelper.executeSQL(connection, "DELETE FROM DOCUMENT_COLLECTION_DOCUMENT_XREF", new Object[0])
            .close();
        DBPersistenceHelper.executeSQL(connection, "DELETE FROM DOCUMENT_COLLECTION", new Object[0]).close();
        DBPersistenceHelper.executeSQL(connection, "DELETE FROM DOCUMENT", new Object[0]).close();
    }

    /**
     * Sets up the test environment. The test instance is obtained. The database tables are cleared. The database
     * connection is open.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    protected void setUp() throws Exception {
        persistence = createPersistence();
        connection = persistence.getConnection();
        connection.setAutoCommit(true);
        clearTables();
    }

    /**
     * Cleans up the test environment. The test instance is disposed. The database tables are cleared. The database
     * connection is closed.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    protected void tearDown() throws Exception {
        clearTables();
        persistence = null;
        connection.close();
        connection = null;
    }

    /**
     * Inserts the delimiters into the DELIMITER table using the given information.
     *
     * @param stringId the string ID of the document which the delimiters belong to.
     * @param delimiters the delimiter to be inserted.
     * @throws SQLException if error occurs when executing the SQL statement.
     */
    protected void insertDelimiters(String stringId, String[] delimiters) throws SQLException {
        for (int i = 0; i < delimiters.length; ++i) {
            DBPersistenceHelper.executeSQL(connection,
                "INSERT INTO DELIMITER" + " (DOCUMENT_ID, DELIMITER_ENTRY) VALUES" + " (?, ?)",
                new Object[] {stringId, delimiters[i]}).close();
        }
    }

    /**
     * Asserts that executing the given SQL would return no row.
     *
     * @param message the message when the assertion fails.
     * @param sql the SQL statement with parameters to be executed.
     * @param params the parameters of the SQL statement to be executed.
     * @throws SQLException if database error occurs when executing the SQL statement.
     */
    protected void assertNoRecord(String message, String sql, Object[] params) throws SQLException {
        PreparedStatement statement = DBPersistenceHelper.executeSQL(connection, sql, params);

        try {
            assertFalse(message, statement.getResultSet().next());
        } finally {
            statement.close();
        }
    }

    /**
     * Tests <code>getDocumentUseCount(WordSourceId)</code> when the document ID is <code>null</code>. Expect
     * IllegalArgumentException.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testGetDocumentUseCountWordSourceIdNull() throws Exception {
        try {
            persistence.getDocumentUseCount(null);
            fail("The document ID is null, should throw IllegalArgumentException.");
        } catch (IllegalArgumentException e) {
            // good
        }
    }

    /**
     * Tests <code>getDocumentUseCount(WordSourceId)</code> when the document source identity instance is not
     * serializable. Expect IndexPersistenceException.
     */
    public void testGetDocumentUseCountWordSourceIdSourceNotSerializable() {

        try {
            persistence.getDocumentUseCount(new WordSourceId("id", "class", new String[] {" "}, Locale.US));
            fail("The document source identity is not serializable, should throw IndexPersistenceException.");
        } catch (IndexPersistenceException e) {
            // good
        }
    }

    /**
     * Tests <code>getDocumentUseCount(WordSourceId)</code> when the document index is not persisted. Expect
     * IndexPersistenceException.
     */
    public void testGetDocumentUseCountWordSourceIdNotPersisted() {
        try {
            persistence.getDocumentUseCount(new WordSourceId("id", "class", new String[] {" ", "\1"}, Locale.US));
            fail("The document index is not persisted, should throw IndexPersistenceException.");
        } catch (IndexPersistenceException e) {
            // good
        }
    }

    /**
     * Tests <code>getDocumentUseCount(WordSourceId)</code> when the document usage count is negative. Expect
     * IndexPersistenceException.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testGetDocumentUseCountWordSourceIdUseCountNegative() throws Exception {
        WordSourceId id = new WordSourceId("id", "class", new String[] {" ", "\1"}, Locale.US);

        // Insert an invalid record
        DBPersistenceHelper.executeSQL(
            connection,
            "INSERT INTO DOCUMENT" + " (DOCUMENT_ID, LOCALE, SOURCE_IDENTITY, CLASS_NAME, USE_COUNT) VALUES"
                + " (?, ?, ?, ?, -1)",
            new Object[] {IndexPersistenceHelper.generateId(id), Locale.US.toString(), new byte[0], "class"}).close();

        try {
            persistence.getDocumentUseCount(id);
            fail("The use count is negative, should throw IndexPersistenceException.");
        } catch (IndexPersistenceException e) {
            // good
        }
    }

    /**
     * Tests accuracy of <code>getDocumentUseCount(WordSourceId)</code>. The retrieved use count should be 5, as
     * stored in the DOCUMENT table.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testGetDocumentUseCountWordSourceIdAccuracy() throws Exception {
        WordSourceId id = new WordSourceId("id", "class", new String[] {" ", "\1"}, Locale.US);

        // Insert a valid record
        DBPersistenceHelper.executeSQL(
            connection,
            "INSERT INTO DOCUMENT" + " (DOCUMENT_ID, LOCALE, SOURCE_IDENTITY, CLASS_NAME, USE_COUNT) VALUES"
                + " (?, ?, ?, ?, 5)",
            new Object[] {IndexPersistenceHelper.generateId(id), Locale.US.toString(), new byte[0], "class"}).close();

        // Verify
        assertEquals("The usage count should be 5.", 5, persistence.getDocumentUseCount(id));
    }

    /**
     * Tests <code>getIndexedDocuments()</code> when a document source identity cannot be deserialized. Expect
     * IndexPersistenceException.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testGetIndexedDocumentsDeserializeError() throws Exception {
        // Create a document id
        WordSourceId id = new WordSourceId("id", "class", new String[] {" ", "\1"}, Locale.US);
        String stringId = IndexPersistenceHelper.generateId(id);

        // Insert an invalid record.
        DBPersistenceHelper.executeSQL(
            connection,
            "INSERT INTO DOCUMENT" + " (DOCUMENT_ID, LOCALE, SOURCE_IDENTITY, CLASS_NAME, USE_COUNT) VALUES"
                + " (?, ?, ?, ?, 5)", new Object[] {stringId, Locale.US.toString(), new byte[] {1}, "class"}).close();

        // Insert the delimiters
        insertDelimiters(stringId, id.getDelimiters());

        try {
            persistence.getIndexedDocuments();
            fail("The document source identity cannot be deserialized, should throw IndexPersistenceException.");
        } catch (IndexPersistenceException e) {
            // good
        }
    }

    /**
     * Tests <code>getIndexedDocuments()</code> when a document ID cannot be created using the data. Expect
     * IndexPersistenceException.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testGetIndexedDocumentsIllegalID() throws Exception {
        // Create a document id
        WordSourceId id = new WordSourceId("id", "class", new String[] {" ", "\1"}, Locale.US);
        String stringId = IndexPersistenceHelper.generateId(id);

        // Insert a valid record, but no delimiter
        DBPersistenceHelper.executeSQL(
            connection,
            "INSERT INTO DOCUMENT" + " (DOCUMENT_ID, LOCALE, SOURCE_IDENTITY, CLASS_NAME, USE_COUNT) VALUES"
                + " (?, ?, ?, ?, 0)",
            new Object[] {stringId, Locale.US.toString(), TestHelper.serialize(id.getSourceIdentity()), "class"})
            .close();

        try {
            persistence.getIndexedDocuments();
            fail("The document ID cannot be created using the data, should throw IndexPersistenceException.");
        } catch (IndexPersistenceException e) {
            // good
        }
    }

    /**
     * Tests accuracy of <code>getIndexedDocuments()</code>. All persisted document index IDs should be retrieved.
     * This is verified by comparing to a reference set of IDs.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testGetIndexedDocumentsAccuracy() throws Exception {
        Set sourceIds = new HashSet();

        // Build the persisted document indices.
        for (int i = 0; i < 5; ++i) {
            WordSourceId id = new WordSourceId("source" + i, "class" + i, new String[] {"delimiter" + i}, Locale.US);
            String stringId = IndexPersistenceHelper.generateId(id);

            sourceIds.add(id);

            // Persist the document ID
            DBPersistenceHelper.executeSQL(
                connection,
                "INSERT INTO DOCUMENT" + " (DOCUMENT_ID, LOCALE, SOURCE_IDENTITY, CLASS_NAME, USE_COUNT) VALUES"
                    + " (?, ?, ?, ?, 0)",
                new Object[] {stringId, id.getSourceLocale().toString(), TestHelper.serialize(id.getSourceIdentity()),
                    id.getSourceClassName()}).close();
            insertDelimiters(stringId, id.getDelimiters());
        }

        // Retrieve the persisted document indices
        Set actual = persistence.getIndexedDocuments();

        // Verify
        assertEquals("All persisted document indices should be retrieved.", sourceIds, actual);
    }

    /**
     * Tests <code>increaseDocumentUseCount(WordSourceId)</code> when the document ID is <code>null</code>. Expect
     * IllegalArgumentException.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testIncreaseDocumentUseCountWordSourceIdNull() throws Exception {
        try {
            persistence.increaseDocumentUseCount(null);
            fail("The document ID is null, should throw IllegalArgumentException.");
        } catch (IllegalArgumentException e) {
            // good
        }
    }

    /**
     * Tests <code>increaseDocumentUseCount(WordSourceId)</code> when the document source identity instance is not
     * serializable. Expect IndexPersistenceException.
     */
    public void testIncreaseDocumentUseCountWordSourceIdSourceNotSerializable() {
        try {
            persistence
                .increaseDocumentUseCount(new WordSourceId(new Object(), "class", new String[] {" "}, Locale.US));
            fail("The document source identity is not serializable, should throw IndexPersistenceException.");
        } catch (IndexPersistenceException e) {
            // good
        }
    }

    /**
     * Tests <code>increaseDocumentUseCount(WordSourceId)</code> when the document index is not persisted. Expect
     * IndexPersistenceException.
     */
    public void testIncreaseDocumentUseCountWordSourceIdNotPersisted() {
        try {
            persistence.increaseDocumentUseCount(new WordSourceId("id", "class", new String[] {" ", "\1"}, Locale.US));
            fail("The document index is not persisted, should throw IndexPersistenceException.");
        } catch (IndexPersistenceException e) {
            // good
        }
    }

    /**
     * Tests <code>increaseDocumentUseCount(WordSourceId)</code> when the document usage count is negative. Expect
     * IndexPersistenceException.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testIncreaseDocumentUseCountWordSourceIdUseCountNegative() throws Exception {
        WordSourceId id = new WordSourceId("id", "class", new String[] {" ", "\1"}, Locale.US);

        // Insert an invalid record
        DBPersistenceHelper.executeSQL(
            connection,
            "INSERT INTO DOCUMENT" + " (DOCUMENT_ID, LOCALE, SOURCE_IDENTITY, CLASS_NAME, USE_COUNT) VALUES"
                + " (?, ?, ?, ?, -1)",
            new Object[] {IndexPersistenceHelper.generateId(id), Locale.US.toString(), new byte[0], "class"}).close();

        try {
            persistence.increaseDocumentUseCount(id);
            fail("The use count is negative, should throw IndexPersistenceException.");
        } catch (IndexPersistenceException e) {
            // good
        }
    }

    /**
     * Tests accuracy of <code>increaseDocumentUseCount(WordSourceId)</code>. The document usage count should be
     * increased by 1. This is verified by get the usage count.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testIncreaseDocumentUseCountWordSourceIdAccuracy() throws Exception {
        WordSourceId id = new WordSourceId("id", "class", new String[] {" ", "\1"}, Locale.US);

        // Insert a valid record
        DBPersistenceHelper.executeSQL(
            connection,
            "INSERT INTO DOCUMENT" + " (DOCUMENT_ID, LOCALE, SOURCE_IDENTITY, CLASS_NAME, USE_COUNT) VALUES"
                + " (?, ?, ?, ?, 5)",
            new Object[] {IndexPersistenceHelper.generateId(id), Locale.US.toString(), new byte[0], "class"}).close();

        // Increase the usage count
        persistence.increaseDocumentUseCount(id);

        // Verify
        assertEquals("The usage count should be increased.", 6, persistence.getDocumentUseCount(id));
    }

    /**
     * Tests <code>decreaseDocumentUseCount(WordSourceId)</code> when the document ID is <code>null</code>. Expect
     * IllegalArgumentException.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testDecreaseDocumentUseCountWordSourceIdNull() throws Exception {
        try {
            persistence.decreaseDocumentUseCount(null);
            fail("The document ID is null, should throw IllegalArgumentException.");
        } catch (IllegalArgumentException e) {
            // good
        }
    }

    /**
     * Tests <code>decreaseDocumentUseCount(WordSourceId)</code> when the document source identity instance is not
     * serializable. Expect IndexPersistenceException.
     */
    public void testDecreaseDocumentUseCountWordSourceIdSourceNotSerializable() {
        try {
            persistence
                .decreaseDocumentUseCount(new WordSourceId(new Object(), "class", new String[] {" "}, Locale.US));
            fail("The document source identity is not serializable, should throw IndexPersistenceException.");
        } catch (IndexPersistenceException e) {
            // good
        }
    }

    /**
     * Tests <code>decreaseDocumentUseCount(WordSourceId)</code> when the document index is not persisted. Expect
     * IndexPersistenceException.
     */
    public void testDecreaseDocumentUseCountWordSourceIdNotPersisted() {
        try {
            persistence.decreaseDocumentUseCount(new WordSourceId("id", "class", new String[] {" ", "\1"}, Locale.US));
            fail("The document index is not persisted, should throw IndexPersistenceException.");
        } catch (IndexPersistenceException e) {
            // good
        }
    }

    /**
     * Tests <code>decreaseDocumentUseCount(WordSourceId)</code> when the document usage count is negative. Expect
     * IndexPersistenceException.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testDecreaseDocumentUseCountWordSourceIdUseCountNegative() throws Exception {
        WordSourceId id = new WordSourceId("id", "class", new String[] {" ", "\1"}, Locale.US);

        // Insert an invalid record
        DBPersistenceHelper.executeSQL(
            connection,
            "INSERT INTO DOCUMENT" + " (DOCUMENT_ID, LOCALE, SOURCE_IDENTITY, CLASS_NAME, USE_COUNT) VALUES"
                + " (?, ?, ?, ?, -1)",
            new Object[] {IndexPersistenceHelper.generateId(id), Locale.US.toString(), new byte[0], "class"}).close();

        try {
            persistence.decreaseDocumentUseCount(id);
            fail("The use count is negative, should throw IndexPersistenceException.");
        } catch (IndexPersistenceException e) {
            // good
        }
    }

    /**
     * Tests <code>decreaseDocumentUseCount(WordSourceId)</code> when the document usage count is 0. Expect
     * IndexPersistenceException.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testDecreaseDocumentUseCountWordSourceIdUseCountZero() throws Exception {
        WordSourceId id = new WordSourceId("id", "class", new String[] {" ", "\1"}, Locale.US);

        // Insert a valid record whose use count is 0
        DBPersistenceHelper.executeSQL(
            connection,
            "INSERT INTO DOCUMENT" + " (DOCUMENT_ID, LOCALE, SOURCE_IDENTITY, CLASS_NAME, USE_COUNT) VALUES"
                + " (?, ?, ?, ?, 0)",
            new Object[] {IndexPersistenceHelper.generateId(id), Locale.US.toString(), new byte[0], "class"}).close();

        try {
            persistence.decreaseDocumentUseCount(id);
            fail("The use count is 0, should throw IndexPersistenceException.");
        } catch (IndexPersistenceException e) {
            // good
        }
    }

    /**
     * Tests accuracy of <code>decreaseDocumentUseCount(WordSourceId)</code>. The document usage count should be
     * decreased by 1. This is verified by get the usage count.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testDecreaseDocumentUseCountWordSourceIdAccuracy() throws Exception {
        WordSourceId id = new WordSourceId("id", "class", new String[] {" ", "\1"}, Locale.US);

        // Insert a valid record
        DBPersistenceHelper.executeSQL(
            connection,
            "INSERT INTO DOCUMENT" + " (DOCUMENT_ID, LOCALE, SOURCE_IDENTITY, CLASS_NAME, USE_COUNT) VALUES"
                + " (?, ?, ?, ?, 5)",
            new Object[] {IndexPersistenceHelper.generateId(id), Locale.US.toString(), new byte[0], "class"}).close();

        // Decrease the usage count
        persistence.decreaseDocumentUseCount(id);

        // Verify
        assertEquals("The usage count should be decreased.", 4, persistence.getDocumentUseCount(id));
    }

    /**
     * Tests <code>addCollectionIndex(CollectionIndex)</code> when the collection index is <code>null</code>.
     * Expect IllegalArgumentException.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testAddCollectionIndexCollectionIndexNull() throws Exception {
        try {
            persistence.addCollectionIndex(null);
            fail("The collection index is null, should throw IllegalArgumentException.");
        } catch (IllegalArgumentException e) {
            // good
        }
    }

    /**
     * Tests <code>addCollectionIndex(CollectionIndex)</code> when the collection index is already persisted. Expect
     * IndexPersistenceException.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testAddCollectionIndexCollectionIndexExist() throws Exception {
        // Create a collection index
        CollectionIndex collection = new CollectionIndex(null, Collections.EMPTY_MAP, Collections.EMPTY_SET, "id");

        // Persist it
        persistence.addCollectionIndex(collection);

        try {
            persistence.addCollectionIndex(collection);
            fail("The collection index is already persisted, should throw IndexPersistenceException.");
        } catch (IndexPersistenceException e) {
            // good
        }
    }

    /**
     * Tests <code>addCollectionIndex(CollectionIndex)</code> when the document index in the collection is not
     * persisted. Expect IndexPersistenceException.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testAddCollectionIndexCollectionIndexDocumentIndexNotExist() throws Exception {
        // Create a document ID
        WordSourceId id = new WordSourceId("id", "class", new String[] {" "}, Locale.US);

        // Create a collection index
        CollectionIndex collection = new CollectionIndex(null, Collections.EMPTY_MAP, Collections.singleton(id), "id");

        try {
            persistence.addCollectionIndex(collection);
            fail("The document index in the collection is not persisted, should throw IndexPersistenceException.");
        } catch (IndexPersistenceException e) {
            // good
        }
    }

    /**
     * Tests <code>addCollectionIndex(CollectionIndex)</code> when a document source identity is not serializable.
     * Expect IndexPersistenceException.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testAddCollectionIndexCollectionIndexSourceNotSerializable() throws Exception {
        WordSourceId id = new WordSourceId(new Object(), "class", new String[] {" "}, Locale.US);

        // Create a collection index
        CollectionIndex collection = new CollectionIndex(null, Collections.EMPTY_MAP, Collections.singleton(id), "id");

        try {
            persistence.addCollectionIndex(collection);
            fail("The document source identity is not serializable, should throw IndexPersistenceException.");
        } catch (IndexPersistenceException e) {
            // good
        }
    }

    /**
     * Tests accuracy of <code>addCollectionIndex(CollectionIndex)</code>. The document string IDs should be
     * persisted in the database. Records should be added to the DOCUMENT_COLLECTION table and the
     * DOCUMENT_COLLECTION_DOCUMENT_XREF table. These are verified by querying the tables.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testAddCollectionIndexCollectionIndexAccuracy() throws Exception {
        Set stringIds = new HashSet();
        Set sourceIds = new HashSet();
        Collator collator = Collator.getInstance(Locale.US);

        collator.setStrength(Collator.PRIMARY);

        // Build the string IDs and document IDs
        for (int i = 0; i < 5; ++i) {
            WordSourceId id = new WordSourceId("source" + i, "class" + i, new String[] {"delimiter" + i}, Locale.US);

            sourceIds.add(id);
            stringIds.add(IndexPersistenceHelper.generateId(id));

            // Create a word map
            Map map = new HashMap();

            for (int j = 0; j < 4; ++j) {
                map.put(collator.getCollationKey("word" + (i / 2 + j)), Collections.singletonList(new Integer(1)));
            }

            // Persist the document index
            persistence.addDocumentIndex(new DocumentIndex(id, map));
        }

        // Persist another document index which is not in the collection index
        persistence.addDocumentIndex(new DocumentIndex(new WordSourceId("no", "class", new String[] {" "}, Locale.US),
            Collections.EMPTY_MAP));

        // Persist the collection index
        persistence.addCollectionIndex(new CollectionIndex(null, Collections.EMPTY_MAP, sourceIds, "id"));

        // Verify DOCUMENT_COLLECTION table, which should have one record with the collection id
        PreparedStatement statement = DBPersistenceHelper.executeSQL(connection,
            "SELECT COUNT(*) FROM DOCUMENT_COLLECTION WHERE COLLECTION_ID = ?", new Object[] {"id"});

        try {
            ResultSet rs = statement.getResultSet();

            rs.next();
            assertEquals("The colleciton index should be persisted.", 1, rs.getInt(1));
        } finally {
            statement.close();
        }

        // Verify DOCUMENT_COLLECTION_DOCUMENT_XREF table.
        statement = DBPersistenceHelper.executeSQL(connection,
            "SELECT DOCUMENT_ID FROM DOCUMENT_COLLECTION_DOCUMENT_XREF WHERE COLLECTION_ID = ?", new Object[] {"id"});

        // Extract all document string IDs in the table belonging to the collection
        Set actual = new HashSet();

        try {
            ResultSet rs = statement.getResultSet();

            while (rs.next()) {
                actual.add(rs.getString(1));
            }
        } finally {
            statement.close();
        }

        assertEquals("All document string IDs should be persisted.", stringIds, actual);
    }

    /**
     * Tests <code>removeCollectionIndex(String)</code> when the collection ID is <code>null</code>. Expect
     * IllegalArgumentException.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testRemoveCollectionIndexStringNull() throws Exception {
        try {
            persistence.removeCollectionIndex(null);
            fail("The collection ID is null, should throw IllegalArgumentException.");
        } catch (IllegalArgumentException e) {
            // good
        }
    }

    /**
     * Tests <code>removeCollectionIndex(String)</code> when the collection ID is empty string. Expect
     * IllegalArgumentException.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testRemoveCollectionIndexStringEmpty() throws Exception {
        try {
            persistence.removeCollectionIndex("  ");
            fail("The collection ID is empty string, should throw IllegalArgumentException.");
        } catch (IllegalArgumentException e) {
            // good
        }
    }

    /**
     * Tests <code>removeCollectionIndex(String)</code> when the collection index does not exist. Expect
     * IndexPersistenceException.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testRemoveCollectionIndexStringNotExist() throws Exception {
        try {
            persistence.removeCollectionIndex("id");
            fail("The collection index does not exist, should throw IndexPersistenceException.");
        } catch (IndexPersistenceException e) {
            // good
        }
    }

    /**
     * Tests accuracy of <code>removeCollectionIndex(String)</code>. The collection index record in the
     * DOCUMENT_COLLECTION table is deleted. All records related to the collection index in the
     * DOCUMENT_COLLECTION_DOCUMENT_XREF table are deleted.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testRemoveCollectionIndexStringAccuracy() throws Exception {
        Set sourceIds = new HashSet();

        // Build the document IDs
        for (int i = 0; i < 5; ++i) {
            WordSourceId id = new WordSourceId("source" + i, "class" + i, new String[] {"delimiter" + i}, Locale.US);

            // Persist the document index
            persistence.addDocumentIndex(new DocumentIndex(id, Collections.EMPTY_MAP));
        }

        // Persist the collection index
        persistence.addCollectionIndex(new CollectionIndex(null, Collections.EMPTY_MAP, sourceIds, "id"));

        // Remove the collection index
        persistence.removeCollectionIndex("id");

        // Verify DOCUMENT_COLLECTION table, which should have no record with the collection id
        assertNoRecord("The colleciton index should be removed.",
            "SELECT * FROM DOCUMENT_COLLECTION WHERE COLLECTION_ID = ?", new Object[] {"id"});

        // Verify DOCUMENT_COLLECTION_DOCUMENT_XREF table.
        assertNoRecord("All document string IDs of this collection should be deleted.",
            "SELECT DOCUMENT_ID FROM DOCUMENT_COLLECTION_DOCUMENT_XREF WHERE COLLECTION_ID = ?", new Object[] {"id"});
    }

    /**
     * Tests <code>updateCollectionIndex(CollectionIndex)</code> when the collection index is <code>null</code>.
     * Expect IllegalArgumentException.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testUpdateCollectionIndexCollectionIndexNull() throws Exception {
        try {
            persistence.updateCollectionIndex(null);
            fail("The collection index is null, should throw IllegalArgumentException.");
        } catch (IllegalArgumentException e) {
            // good
        }
    }

    /**
     * Tests <code>updateCollectionIndex(CollectionIndex)</code> when the collection index is not persisted. Expect
     * IndexPersistenceException.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testUpdateCollectionIndexCollectionIndexNotExist() throws Exception {
        // Create a collection index
        CollectionIndex collection = new CollectionIndex(null, Collections.EMPTY_MAP, Collections.EMPTY_SET, "id");

        try {
            persistence.updateCollectionIndex(collection);
            fail("The collection index is not persisted, should throw IndexPersistenceException.");
        } catch (IndexPersistenceException e) {
            // good
        }
    }

    /**
     * Tests <code>updateCollectionIndex(CollectionIndex)</code> when the document index in the collection is not
     * persisted. Expect IndexPersistenceException.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testUpdateCollectionIndexCollectionIndexDocumentIndexNotExist() throws Exception {
        // Create a document ID
        WordSourceId id = new WordSourceId("id", "class", new String[] {" "}, Locale.US);

        // Create and persist the collection index
        persistence.addCollectionIndex(new CollectionIndex(null, Collections.EMPTY_MAP, Collections.EMPTY_SET, "id"));

        // Create the same collection index
        CollectionIndex collection = new CollectionIndex(null, Collections.EMPTY_MAP, Collections.singleton(id), "id");

        try {
            persistence.updateCollectionIndex(collection);
            fail("The document index in the collection is not persisted, should throw IndexPersistenceException.");
        } catch (IndexPersistenceException e) {
            // good
        }
    }

    /**
     * Tests <code>updateCollectionIndex(CollectionIndex)</code> when a document source identity is not serializable.
     * Expect IndexPersistenceException.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testUpdatecollectionIndexCollectionIndexSourceNotSerializable() throws Exception {
        WordSourceId id = new WordSourceId(new Object(), "class", new String[] {" "}, Locale.US);

        // Create and persist the collection index
        persistence.addCollectionIndex(new CollectionIndex(null, Collections.EMPTY_MAP, Collections.EMPTY_SET, "id"));

        // Create the same collection index
        CollectionIndex collection = new CollectionIndex(null, Collections.EMPTY_MAP, Collections.singleton(id), "id");

        try {
            persistence.updateCollectionIndex(collection);
            fail("The document source identity is not serializable, should throw IndexPersistenceException.");
        } catch (IndexPersistenceException e) {
            // good
        }
    }

    /**
     * Tests accuracy of <code>updateCollectionIndex(CollectionIndex)</code>. The document string IDs should be
     * persisted in the database. The DOCUMENT_COLLECTION_DOCUMENT_XREF table should be updated. This is verified by
     * querying the tables.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testUpdateCollectionIndexCollectionIndexAccuracy() throws Exception {
        Set stringIds = new HashSet();
        Set sourceIds = new HashSet();
        Collator collator = Collator.getInstance(Locale.US);

        collator.setStrength(Collator.PRIMARY);

        // Build the string IDs and document IDs to be updated
        for (int i = 0; i < 5; ++i) {
            WordSourceId id = new WordSourceId("source" + i, "class" + i, new String[] {"delimiter" + i}, Locale.US);

            sourceIds.add(id);
            stringIds.add(IndexPersistenceHelper.generateId(id));

            // Create a word map
            Map map = new HashMap();

            for (int j = 0; j < 4; ++j) {
                map.put(collator.getCollationKey("word" + (i / 2 + j)), Collections.singletonList(new Integer(1)));
            }

            // Persist the document index
            persistence.addDocumentIndex(new DocumentIndex(id, map));
        }

        // Persist another document index which is initially in the collection index
        WordSourceId singleId = new WordSourceId("no", "class", new String[] {" "}, Locale.US);

        persistence.addDocumentIndex(new DocumentIndex(new WordSourceId("no", "class", new String[] {" "}, Locale.US),
            Collections.EMPTY_MAP));

        // Persist the initial collection index
        persistence.addCollectionIndex(new CollectionIndex(null, Collections.EMPTY_MAP,
            Collections.singleton(singleId), "id"));

        // Update the collection index
        persistence.updateCollectionIndex(new CollectionIndex(null, Collections.EMPTY_MAP, sourceIds, "id"));

        // Verify DOCUMENT_COLLECTION_DOCUMENT_XREF table.
        PreparedStatement statement = DBPersistenceHelper.executeSQL(connection,
            "SELECT DOCUMENT_ID FROM DOCUMENT_COLLECTION_DOCUMENT_XREF WHERE COLLECTION_ID = ?", new Object[] {"id"});

        // Extract all document string IDs in the table belonging to the collection
        Set actual = new HashSet();

        try {
            ResultSet rs = statement.getResultSet();

            while (rs.next()) {
                actual.add(rs.getString(1));
            }
        } finally {
            statement.close();
        }

        assertEquals("All document string IDs should be persisted.", stringIds, actual);
    }

    /**
     * Tests <code>removeDocumentIndex(WordSourceId)</code> when the document ID is <code>null</code>. Expect
     * IllegalArgumentException.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testRemoveDocumentIndexWordSourceIdNull() throws Exception {
        try {
            persistence.removeDocumentIndex(null);
            fail("The document ID is null, should throw IllegalArgumentException.");
        } catch (IllegalArgumentException e) {
            // good
        }
    }

    /**
     * Tests <code>removeDocumentIndex(WordSourceId)</code> when the document source identity instance is not
     * serializable. Expect IndexPersistenceException.
     */
    public void testRemoveDocumentIndexWordSourceIdSourceNotSerializable() {
        try {
            persistence.removeDocumentIndex(new WordSourceId(new Object(), "class", new String[] {" "}, Locale.US));
            fail("The document source identity is not serializable, should throw IndexPersistenceException.");
        } catch (IndexPersistenceException e) {
            // good
        }
    }

    /**
     * Tests <code>removeDocumentIndex(WordSourceId)</code> when the document index is not persisted. Expect
     * IndexPersistenceException.
     */
    public void testRemoveDocumentIndexWordSourceIdNotPersisted() {
        try {
            persistence.removeDocumentIndex(new WordSourceId("id", "class", new String[] {" ", "\1"}, Locale.US));
            fail("The document index is not persisted, should throw IndexPersistenceException.");
        } catch (IndexPersistenceException e) {
            // good
        }
    }

    /**
     * Tests <code>removeDocumentIndex(WordSourceId)</code> when the usage count of the document index is 0. Expect
     * IndexPersistenceException.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testRemoveDocumentIndexWordSourceIdUsageCountNotZero() throws Exception {
        WordSourceId id = new WordSourceId("id", "class", new String[] {" ", "\1"}, Locale.US);

        // Persist the document index with no words
        persistence.addDocumentIndex(new DocumentIndex(id, Collections.EMPTY_MAP));

        // Increase the usage count
        persistence.increaseDocumentUseCount(id);

        try {
            persistence.removeDocumentIndex(id);
            fail("The usage count of the document index is not 0, should throw IndexPersistenceException.");
        } catch (IndexPersistenceException e) {
            // good
        }
    }

    /**
     * Tests <code>getCollectionIndex(String)</code> when the collection ID is <code>null</code>. Expect
     * IllegalArgumentException.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testGetCollectionIndexStringNull() throws Exception {
        try {
            persistence.getCollectionIndex(null);
            fail("The collection ID is null, should throw IllegalArgumentException.");
        } catch (IllegalArgumentException e) {
            // good
        }
    }

    /**
     * Tests <code>getCollectionIndex(String)</code> when the collection ID is empty string. Expect
     * IllegalArgumentException.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testGetCollectionIndexStringEmpty() throws Exception {
        try {
            persistence.getCollectionIndex("  ");
            fail("The collection ID is empty string, should throw IllegalArgumentException.");
        } catch (IllegalArgumentException e) {
            // good
        }
    }

    /**
     * Tests <code>getCollectionIndex(String)</code> when a document source identity cannot be deserialized. Expect
     * IndexPersistenceException.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testGetCollectionIndexStringDeserializeError() throws Exception {
        WordSourceId id = new WordSourceId("id", "class", new String[] {" ", "\1"}, Locale.US);
        String stringId = IndexPersistenceHelper.generateId(id);

        // Add a document index with no words
        persistence.addDocumentIndex(new DocumentIndex(id, Collections.EMPTY_MAP));

        // Add a collection index
        persistence
            .addCollectionIndex(new CollectionIndex(null, Collections.EMPTY_MAP, Collections.singleton(id), "id"));

        // Update the document index, so that the source identity cannot be deserialized
        DBPersistenceHelper.executeSQL(connection, "UPDATE DOCUMENT SET SOURCE_IDENTITY = ? WHERE DOCUMENT_ID = ?",
            new Object[] {new byte[] {1, 2, 3, 4, 5}, stringId}).close();

        try {
            persistence.getCollectionIndex("id");
            fail("The document source identity cannot be deserialized, should throw IndexPersistenceException.");
        } catch (IndexPersistenceException e) {
            // good
        }
    }

    /**
     * Tests <code>getCollectionIndex(String)</code> when a document ID cannot be created using the data. Expect
     * IndexPersistenceException.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testGetCollectionIndexStringIllegalID() throws Exception {
        // Create a document id
        WordSourceId id = new WordSourceId("id", "class", new String[] {" ", "\1"}, Locale.US);
        String stringId = IndexPersistenceHelper.generateId(id);

        // Persist the document index
        persistence.addDocumentIndex(new DocumentIndex(id, Collections.EMPTY_MAP));

        // Persist the collection index
        persistence
            .addCollectionIndex(new CollectionIndex(null, Collections.EMPTY_MAP, Collections.singleton(id), "id"));

        // Delete all delimiters in the document index
        DBPersistenceHelper.executeSQL(connection, "DELETE FROM DELIMITER WHERE DOCUMENT_ID = ?",
            new Object[] {stringId}).close();

        try {
            persistence.getCollectionIndex("id");
            fail("The document ID cannot be created using the data, should throw IndexPersistenceException.");
        } catch (IndexPersistenceException e) {
            // good
        }
    }

    /**
     * Tests accuracy of <code>getCollectionIndex(String)</code> when the collection index is not persisted. It should
     * return <code>null</code>.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testGetCollectionIndexStringNotExistAccuracy() throws Exception {
        assertNull("The document index is not persisted, should return null.", persistence.getCollectionIndex("id"));
    }

    /**
     * Tests accuracy of <code>getCollectionIndex(String)</code> when the collection index is persisted. The
     * collection index should be retrieved from the persistence. The document IDs should be read back. The word map
     * should be built according to the word maps in the document indices. These are verified by comparing the
     * collection index to reference instances.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testGetCollectionIndexStringExistAccuracy() throws Exception {
        Set sourceIds = new HashSet();
        Collator collator = Collator.getInstance(Locale.US);
        Map expected = new HashMap();

        collator.setStrength(Collator.PRIMARY);

        // Build the document IDs, and the expected word map
        for (int i = 0; i < 5; ++i) {
            WordSourceId id = new WordSourceId("source" + i, "class" + i, new String[] {"delimiter" + i}, Locale.US);

            sourceIds.add(id);

            // Create a word map
            Map map = new HashMap();

            for (int j = 0; j < 4; ++j) {
                map.put(collator.getCollationKey("word" + (i / 2 + j)), Collections.singletonList(new Integer(1)));
            }

            DocumentIndex index = new DocumentIndex(id, map);

            // Add document index persistence
            persistence.addDocumentIndex(index);

            // Build the expected word map
            IndexPersistenceHelper.addDocumentWordMap(expected, index);
        }

        // Persist the collection index
        persistence.addCollectionIndex(new CollectionIndex(null, expected, sourceIds, "id"));

        // Get the collection index
        CollectionIndex collection = persistence.getCollectionIndex("id");

        // Verify the word map and the document IDs
        assertEquals("The word map should be the same as reference.", expected, collection.getWords());
        assertEquals("The document IDs should be the same as reference.", sourceIds, collection.getAllDocumentIds());
    }

    /**
     * Tests <code>addDocumentIndex(DocumentIndex)</code> when the document index is <code>null</code>. Expect
     * IllegalArgumentException.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testAddDocumentIndexDocumentIndexNull() throws Exception {
        try {
            persistence.addDocumentIndex(null);
            fail("The document index is null, should throw IllegalArgumentException.");
        } catch (IllegalArgumentException e) {
            // good
        }
    }

    /**
     * Tests <code>addDocumentIndex(DocumentIndex)</code> when the document source identity instance is not
     * serializable. Expect IndexPersistenceException.
     */
    public void testAddDocumentIndexDocumentIndexSourceNotSerializable() {
        try {
            persistence.addDocumentIndex(new DocumentIndex(new WordSourceId(new Object(), "class", new String[] {" "},
                Locale.US), Collections.EMPTY_MAP));
            fail("The document source identity is not serializable, should throw IndexPersistenceException.");
        } catch (IndexPersistenceException e) {
            // good
        }
    }

    /**
     * Tests <code>addDocumentIndex(DocumentIndex)</code> when the document index is already persisted. Expect
     * IndexPersistenceException.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testAddDocumentIndexDocumentIndexPersisted() throws Exception {
        DocumentIndex index = new DocumentIndex(new WordSourceId("id", "class", new String[] {" "}, Locale.US),
            Collections.EMPTY_MAP);

        // Persist the document index
        persistence.addDocumentIndex(index);

        try {
            // Persist it again
            persistence.addDocumentIndex(index);
            fail("The document index is already persisted, should throw IndexPersistenceException.");
        } catch (IndexPersistenceException e) {
            // good
        }
    }

    /**
     * Tests <code>getDocumentIndex(WordSourceId)</code> when the document ID is <code>null</code>. Expect
     * IllegalArgumentException.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testGetDocumentIndexWordSourceIdNull() throws Exception {
        try {
            persistence.getCollectionIndex(null);
            fail("The document ID is null, should throw IllegalArgumentException.");
        } catch (IllegalArgumentException e) {
            // good
        }
    }

    /**
     * Tests <code>getDocumentIndex(WordSourceId)</code> when the document source identity instance is not
     * serializable. Expect IndexPersistenceException.
     */
    public void testGetDocumentIndexWordSourceIdSourceNotSerializable() {
        try {
            persistence.getDocumentIndex(new WordSourceId(new Object(), "class", new String[] {" "}, Locale.US));
            fail("The document source identity is not serializable, should throw IndexPersistenceException.");
        } catch (IndexPersistenceException e) {
            // good
        }
    }

    /**
     * Tests accuracy of <code>getDocumentIndex(WordSourceId)</code> when the document index is not persisted. It
     * should return <code>null</code>.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testGetDocumentIndexWordSourceIdNotPersistedAccuracy() throws Exception {
        assertNull("The document index is not persisted, should return null.", persistence
            .getDocumentIndex(new WordSourceId("id", "class", new String[] {" ", "\1"}, Locale.US)));
    }

    /**
     * Tests accuracy of <code>getDocumentIndex(WordSourceId)</code>. The word map of the loaded document index
     * should be exactly the same as a reference word map.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testGetDocumentIndexWordSourceIdAccuracy() throws Exception {
        // Create a document id
        WordSourceId id = new WordSourceId("id", "class", new String[] {" ", "\1"}, Locale.US);

        // Create the expected word map
        Map wordMap = new HashMap();
        Collator collator = Collator.getInstance(Locale.US);

        collator.setStrength(Collator.PRIMARY);

        for (int i = 0; i < 6; ++i) {
            List positions = new ArrayList();

            for (int j = 0; j < 10; ++j) {
                positions.add(new Integer(j));
            }

            wordMap.put(collator.getCollationKey("Word" + Integer.toString(i)), positions);
        }

        // Add another word containing illegal XML character
        wordMap.put(collator.getCollationKey("\1"), Collections.singletonList(new Integer(0)));

        // Persist the document index
        persistence.addDocumentIndex(new DocumentIndex(id, wordMap));

        // Load the document index
        DocumentIndex actual = persistence.getDocumentIndex(id);

        // Verify
        assertEquals("The loaded word map should be the same as the reference one.", wordMap, actual.getWords());
    }
}





